@@ -9,6 +9,7 @@ import torch.nn as nn
 
 from scipy import sparse
 from scipy.stats import norm
+from torch import Tensor
 from tqdm import tqdm
 
 import numpy as np
@@ -575,74 +576,13 @@ def binary_indicator(labels_true: collection, labels_pred: collection) -> Tuple[
     return acc_s, rec_s, f1_s, fpr, tpr, auroc_s, auprc_s
 
 
-class RandomWalkModel(nn.Module):
-    def __init__(self, gamma: float, epsilon: float, p: int):
-        super().__init__()
-        self.gamma = gamma
-        self.epsilon = epsilon
-        self.p = p
-
-    def forward(self, weight, pt, p0):
-        factor = 1 - self.gamma
-        delta = 1.0
-        k = 0
-
-        while delta > self.epsilon:
-            p1 = factor * torch.matmul(weight, pt) + self.gamma * p0
-            delta = torch.linalg.norm(pt - p1, ord=self.p).item()
-            pt = p1
-            k += 1
-
-        return pt
-
-
-def random_walk_parallel(seed_cell_vector, weight, gamma: float = 0.05, epsilon: float = 1e-5, p: int = 2, device='auto'):
-
-    is_gpu_available = check_gpu_availability(verbose=False)
-
-    if device == 'auto' and is_gpu_available:
-        device = 'cuda'
-    else:
-        device = 'cpu'
-
-    weight = to_dense(weight, is_array=True)
-
-    weight = torch.as_tensor(weight, device=device, dtype=torch.float32)
-    p0 = torch.as_tensor(seed_cell_vector, device=device, dtype=torch.float32).unsqueeze(1)
-    pt = p0.clone()
-
-    model = RandomWalkModel(gamma, epsilon, p).to(device)
-
-    if device == 'cuda' and torch.cuda.device_count() > 1:
-
-        class CustomDataParallel(nn.DataParallel):
-
-            def scatter(self, inputs, kwargs, device_ids):
-                # 只在第一个参数（weight）的 0 维度上分割
-                _weight_, _pt_, _p0_ = inputs
-                scattered_weight = torch.nn.parallel.scatter(_weight_, device_ids, dim=0)
-                # 对 pt 和 p0 不进行分割，而是复制到所有设备
-                scattered_pt = [_pt_.to(_device_) for _device_ in device_ids]
-                scattered_p0 = [_p0_.to(_device_) for _device_ in device_ids]
-                # 重新组合输入
-                scattered_inputs = [(sw, sp, sp0) for sw, sp, sp0 in zip(scattered_weight, scattered_pt, scattered_p0)]
-                return scattered_inputs, [kwargs] * len(device_ids)
-
-        model = CustomDataParallel(model)
-
-    with torch.no_grad():
-        result = model(weight, pt, p0)
-
-    return result.cpu().numpy().flatten()
-
-
-def random_walk_single(
+def _random_walk_cpu_(
     seed_cell_vector: collection,
     weight: matrix_data = None,
     gamma: float = 0.05,
     epsilon: float = 1e-5,
     p: int = 2
-) -> matrix_data:
+) -> collection:
     """
     Perform a random walk
     :param seed_cell_vector: seed cells;
@@ -679,6 +619,107 @@ def random_walk_single(
     return pt.flatten()
 
 
+class RandomWalkModel(nn.Module):
+
+    def __init__(self, weight: Union[matrix_data, Tensor], gamma: float = 0.05, epsilon: float = 1e-5, p: int = 2, device: str = 'auto', pbar=None):
+        super().__init__()
+        self.gamma = gamma
+        self.epsilon = epsilon
+        self.p = p
+        self.pbar = pbar
+
+        is_gpu_available = check_gpu_availability(verbose=False)
+
+        self.device = 'cuda' if (device == 'gpu' or (device == 'auto' and is_gpu_available)) else 'cpu'
+
+        self.weight = torch.as_tensor(to_dense(weight, is_array=True), device=self.device, dtype=torch.float32) if isinstance(weight, matrix_data) else weight.to(self.device)
+
+        self.factor = 1 - self.gamma
+
+    def core(self, seed_cell_vector: Tensor):
+
+        p0 = seed_cell_vector
+        pt = p0.clone()
+
+        delta = 1.0
+        k = 0
+
+        while delta > self.epsilon:
+            p1 = self.factor * torch.matmul(self.weight, pt) + self.gamma * p0
+            delta = torch.linalg.norm(pt - p1, ord=self.p).item()
+            pt = p1
+            k += 1
+
+        if self.pbar is not None:
+            self.pbar.update(1)
+
+        return pt.flatten()
+
+    def forward(self, seed_cell_weight: Tensor):
+
+        sample_count = seed_cell_weight.shape[1]
+
+        score = torch.zeros(seed_cell_weight.shape)
+
+        for i in range(sample_count):
+            score[:, i] = self.core(seed_cell_weight[:, i])
+
+        return score
+
+
+def _random_walk_gpu_(
+    seed_cell_weight: matrix_data,
+    weight: matrix_data,
+    gamma: float = 0.05,
+    epsilon: float = 1e-5,
+    p: int = 2,
+    device: str = 'auto'
+) -> matrix_data:
+
+    with tqdm(total=seed_cell_weight.shape[1]) as pbar:
+
+        model = RandomWalkModel(weight, gamma, epsilon, p, device, pbar)
+
+        device = model.device
+        model.to(device)
+
+        seed_cell_weight = torch.as_tensor(seed_cell_weight, device=device, dtype=torch.float32)
+
+        if device == 'cuda' and torch.cuda.device_count() < seed_cell_weight.shape[1]:
+            model = nn.DataParallel(model, dim=1)
+
+        with torch.no_grad():
+            result = model(seed_cell_weight)
+
+        return result.cpu().numpy().flatten()
+
+
+def random_walk(
+    seed_cell_weight: matrix_data,
+    weight: matrix_data,
+    gamma: float = 0.05,
+    epsilon: float = 1e-5,
+    p: int = 2,
+    device: str = 'auto'
+) -> matrix_data:
+
+    availability = check_gpu_availability(False)
+
+    if device == 'cpu' or (device == 'auto' and not availability):
+        sample_count = seed_cell_weight.shape[1]
+
+        score = np.zeros(seed_cell_weight.shape)
+
+        for i in tqdm(range(sample_count)):
+            score[:, i] = _random_walk_cpu_(seed_cell_weight[:, i], weight, gamma, epsilon, p)
+
+        return score
+    elif device == 'gpu' or (device == 'auto' and availability):
+        return _random_walk_gpu_(seed_cell_weight, weight, gamma, epsilon, p, device='gpu')
+    else:
+        raise ValueError(f'The `device` ({device}) is not supported. Only supports "cpu", "gpu", and "auto" values.')
+
+
 class RandomWalk:
     """
     Random walk
@@ -883,10 +924,10 @@ class RandomWalk:
                 self.seed_cell_weight_en_ncw
             ) = self._get_seed_cell_(init_data=init_status_no_weight, info="ablation")
 
-    def _random_walk_(self, seed_cell_vector: collection, weight: matrix_data = None, gamma: float = 0.05, device: str = 'auto') -> matrix_data:
+    def _random_walk_(self, seed_cell_data: matrix_data, weight: matrix_data = None, gamma: float = 0.05, device: str = 'auto') -> matrix_data:
         """
         Perform a random walk
-        :param seed_cell_vector: seed cells;
+        :param seed_cell_data: seed cells;
         :param weight: weight matrix;
         :param gamma: reset weight.
         :param device: device.
@@ -899,24 +940,24 @@ class RandomWalk:
             w = weight
 
         if not self.is_gpu_available:
-            return random_walk_single(seed_cell_vector, weight=w, gamma=gamma, epsilon=self.epsilon, p=self.p)
+            return random_walk(seed_cell_data, weight=w, gamma=gamma, epsilon=self.epsilon, p=self.p, device='cpu')
 
         try:
-            _data_ = random_walk_parallel(seed_cell_vector, weight=w, gamma=gamma, epsilon=self.epsilon, p=self.p, device=device)
+            _data_ = random_walk(seed_cell_data, weight=w, gamma=gamma, epsilon=self.epsilon, p=self.p, device=device)
         except Exception as e:
             ul.log(__name__).error(f"GPU failed to run, try to switch to CPU running.\n {e}")
-            _data_ = random_walk_single(seed_cell_vector, weight=w, gamma=gamma, epsilon=self.epsilon, p=self.p)
+            _data_ = random_walk(seed_cell_data, weight=w, gamma=gamma, epsilon=self.epsilon, p=self.p, device='cpu')
 
         return _data_
 
-    def _random_walk_core_(self, seed_cell_vector: collection, weight: matrix_data = None) -> matrix_data:
+    def _random_walk_core_(self, seed_cell_data: matrix_data, weight: matrix_data = None) -> matrix_data:
         """
         Perform a random walk
-        :param seed_cell_vector: seed cells;
+        :param seed_cell_data: seed cells;
         :param weight: weight matrix.
         :return: The value after random walk.
         """
-        return self._random_walk_(seed_cell_vector, weight, self.gamma)
+        return self._random_walk_(seed_cell_data, weight, self.gamma)
 
     @staticmethod
     def _get_weight_(cell_cell_matrix: matrix_data) -> matrix_data:
@@ -1143,7 +1184,7 @@ class RandomWalk:
         with tqdm(total=total_steps) as pbar:
             for i in self.trait_range:
 
-                cell_value = np.zeros(self.cell_size)
+                random_seed_cell_matrix = np.zeros((self.cell_size, self.benchmark_count))
 
                 for _ in range(self.benchmark_count):
                     # Set random seed information
@@ -1161,12 +1202,14 @@ class RandomWalk:
                     if trait_value_min != trait_value_max:
                         # seed cell weight
                         random_seed_cell[random_seed_index] = 1 / self.cell_size
+                        random_seed_cell_matrix = random_seed_cell
 
-                        # Random walk
-                        cell_value += self._random_walk_core_(random_seed_cell)
                     pbar.update(1)
+
+                # Random walk
+                cell_value_matrix = self._random_walk_core_(random_seed_cell_matrix)
                 # Remove the influence of background
-                self.random_seed_cell[:, i] = cell_value / self.benchmark_count
+                self.random_seed_cell[:, i] = cell_value_matrix.mean(axis=0)
 
         cell_value = self.scale_norm(self.random_seed_cell)
         self.trs_adata.layers["benchmark"] = to_sparse(cell_value)
@@ -1205,13 +1248,10 @@ class RandomWalk:
         if weight is None:
             weight = self.weight
 
-        score = np.zeros(self.trs_adata.shape)
-
         _log_info_, _layer_label_ = self._get_label_description_(label)
         ul.log(__name__).info(f"Calculate {len(self.trait_list)} traits/diseases for process `{label}`. ({_log_info_} ===> `{_layer_label_}`)")
 
-        for i in tqdm(self.trait_range):
-            score[:, i] = self._random_walk_core_(seed_cell_data[:, i], weight=weight)
+        score = self._random_walk_core_(seed_cell_data, weight=weight)
 
         cell_value = self.scale_norm(score)
 
@@ -1346,15 +1386,19 @@ class RandomWalk:
         trait_cell_enrichment = np.zeros(self.trs_adata.shape)
         trait_cell_credible = np.zeros(self.trs_adata.shape)
 
-        ul.log(__name__).info(f"Calculate {len(self.trait_list)} traits/diseases for process `{label}`. (Enrichment)")
+        ul.log(__name__).info(f"Calculate {len(self.trait_list)} traits/diseases for process `{label}`. (Enrichment-random walk)")
+        # Random walk
+        cell_value_data = self._random_walk_(
+            seed_cell_en_weight,
+            weight=self.weight_m_knn if label == "run_en_ablation_m_knn" else self.weight,
+            gamma=self.enrichment_gamma
+        )
+
+        ul.log(__name__).info(f"Calculate {len(self.trait_list)} traits/diseases for process `{label}`. (Enrichment-score)")
         for i in tqdm(self.trait_range):
 
             # Random walk
-            cell_value = self._random_walk_(
-                seed_cell_en_weight[:, i],
-                weight=self.weight_m_knn if label == "run_en_ablation_m_knn" else self.weight,
-                gamma=self.enrichment_gamma
-            )
+            cell_value = cell_value_data[:, i]
 
             # separate
             cell_value_credible = mean_symmetric_scale(np.array(source_value[:, i]).flatten() - np.array(cell_value).flatten(), is_verbose=False)
